@@ -22,10 +22,6 @@ from fabric import colors
 
 from raptor.models import Proc
 
-# TODO: Change these to /apps/ so that we don't have YG-isms in the code and
-# can more safely bind-mount /opt into app environments.  It's tricky, because
-# cleanup code will have to check both the old and new locations when it runs.
-# OR we do a one-time migration.
 PROCS_ROOT = '/apps/procs'
 RELEASES_ROOT = '/apps/releases'
 
@@ -38,8 +34,6 @@ LEGACY_RELEASES_ROOT = '/opt/yg/releases'
 
 @task
 def upload_release(build_path, config_path, release_path, envsh_path, user='nobody'):
-    # TODO: accept env_path that points to an env.sh file that can be sourced
-    # by the proc at startup to provide env vars.
 
     remote_procfile = posixpath.join(release_path, 'Procfile')
 
@@ -95,14 +89,17 @@ def get_template(name):
     return os.path.join(here, 'templates', name)
 
 
-class Deployment(object):
-    def __init__(self, release_name, proc, port, user, use_syslog, contain):
+class Deployer(object):
+    mountpoints = []
+    lxc_config_template = None
+    start_script_template = 'start_proc.sh'
+
+    def __init__(self, release_name, proc, port, user, use_syslog):
         self.release_name = release_name
         self.proc = proc
         self.port = port
         self.user = user
         self.use_syslog = use_syslog
-        self.contain = contain
         self.proc_name = '-'.join([release_name, proc, str(port)])
         self.proc_path = posixpath.join(PROCS_ROOT, self.proc_name)
         self.proc_tmp_path = posixpath.join(self.proc_path, 'tmp')
@@ -112,29 +109,51 @@ class Deployment(object):
                                                  "settings.yaml")
         self.proc_line = self.get_proc_line()
 
-    def run(self):
-
+    def configure_proc(self):
         sudo('mkdir -p ' + self.proc_path)
 
         self.write_proc_conf()
+        self.write_proc_lxc()
         self.write_start_proc_sh()
         self.create_proc_tmpdir()
         self.upload_uptester()
 
-        if self.contain:
-            # Make container mount points.  lxc will handle the actual mounts,
-            # but we need to make the mount points.
-            mountpoints = ('/app', '/bin', '/dev', '/etc', '/lib', '/lib64',
-                           '/opt', '/usr', '/proc', '/sys', '/dev/pts',
-                           '/dev/shm')
-            for m in mountpoints:
-                sudo('mkdir -p %s%s' % (self.proc_path, m))
+        self.create_mount_points()
 
+    def run(self):
+        self.configure_proc()
+        self.reload_supervisor()
+
+    def get_paths(self):
+        """
+        Return a dictionary of paths to be used when writing the proc config.
+        Must include keys for 'home', 'tmp', 'settings', and 'envsh'.
+        """
+        return {
+            'home': self.proc_tmp_path,
+            'tmp': self.proc_tmp_path,
+            'settings': self.yaml_settings_path,
+            'envsh': self.envsh_path,
+        }
+
+    def get_start_cmd(self):
+        """
+        Return the command that should be run to actually start up the process.
+        """
+        return 'exec %s' % self.proc_line
+
+    def reload_supervisor(self):
         # For this to work, the host must have PROCS_ROOT/*/proc.conf in
         # the files include line in the main supervisord.conf
         sudo('supervisorctl reread')
         sudo('supervisorctl add ' + self.proc_name)
 
+    def create_mount_points(self):
+        # Make container mount points.  lxc will handle the actual mounts,
+        # but we need to make the folders.
+        for m in self.mountpoints:
+            sudo('mkdir -p %s%s' % (self.proc_path, m))
+
     def write_proc_conf(self):
         if self.use_syslog:
             stdout_log = "syslog"
@@ -146,7 +165,7 @@ class Deployment(object):
             'port': self.port,
             'proc_path': self.proc_path,
             'stdout_log': stdout_log,
-            'user': 'root' if self.contain else self.user,
+            'user': self.get_proc_conf_user(),
             'release_path': self.release_path,
         }
         proc_conf_tmpl = get_template('proc.conf')
@@ -154,32 +173,36 @@ class Deployment(object):
         files.upload_template(proc_conf_tmpl, remote_supd, proc_conf_vars,
                               use_sudo=True)
 
-    def write_start_proc_sh(self):
-        if self.contain:
-            sh_script = get_template('start_contained.sh')
-            # write the lxc config.
-            lxc_tmpl = get_template('proc.lxc')
+    def write_proc_lxc(self):
+        # write the lxc config.
+        if self.lxc_config_template:
+            lxc_tmpl = get_template(self.lxc_config_template)
             remote_lxc_path = posixpath.join(self.proc_path, 'proc.lxc')
             files.upload_template(lxc_tmpl, remote_lxc_path, {
                 'proc_path': self.proc_path,
                 'release_path': self.release_path,
             }, use_sudo=True)
-        else:
-            sh_script = get_template('start_proc.sh')
+
+    def write_start_proc_sh(self):
+        sh_script = get_template('start_proc.sh')
 
         sh_remote = posixpath.join(self.proc_path, 'start_proc.sh')
-        files.upload_template(sh_script, sh_remote, {
-            'envsh_path': '/app/env.sh' if self.contain else self.envsh_path,
-            'settings_path': self.release_path + "/settings.yaml",
-            'port': self.port,
-            'cmd': self.proc_line,
-            'tmpdir': self.proc_tmp_path,
-            'proc_name': self.proc_name,
-            'procs_root': PROCS_ROOT,
-            'user': self.user,
-        }, use_sudo=True)
+
+        # The start_proc.sh template needs to be fed some paths, as well as a
+        # port and the command to be launched
+        tmpl_data = self.get_paths()
+        tmpl_data['cmd'] = self.get_start_cmd()
+        tmpl_data['port'] = self.port
+
+        files.upload_template(sh_script, sh_remote, tmpl_data, use_sudo=True)
         sudo('chmod +x %s' % sh_remote)
 
+    def get_proc_conf_user(self):
+        # Uncontained procs are just executed by Supervisor directly, so
+        # proc.conf needs to be configured with the real username that we want
+        # to run as
+        return self.user
+
     def create_proc_tmpdir(self):
         # Create a place for the proc to stick temporary files if it needs to.
         sudo('mkdir -p ' + self.proc_tmp_path)
@@ -217,12 +240,94 @@ class Deployment(object):
         sudo('chmod +x %s' % remote_path)
 
 
+class ContainedDeployer(Deployer):
+    start_script_template = 'start_contained.sh'
+
+    def get_proc_conf_user(self):
+        # Contained procs will be started by Supervisor as root, and then su to
+        # be the configured user once inside the container.
+        return 'root'
+
+    def get_paths(self):
+        """
+        Return a dictionary of paths to be used when writing the proc config.
+        Must include keys for 'home', 'tmp', 'settings', and 'envsh'.
+        """
+        # In a container, these paths are always the same.
+        return {
+            'home': '/tmp',
+            'tmp': '/tmp',
+            'settings': '/app/settings.yaml',
+            'envsh': '/app/env.sh',
+        }
+
+    def get_start_cmd(self):
+        inner_cmd = self.proc_line
+        container_name = self.proc_name
+        lxc_config_path = posixpath.join(self.proc_path, 'proc.lxc')
+        return build_container_cmd(inner_cmd, self.user, container_name,
+                                   lxc_config_path)
+
+
+def build_container_cmd(cmd, user, container_name, lxc_config_path):
+    """
+    Build up the shell command needed to launch a program inside a container.
+    This is used both in deployments and launching the uptester.
+    """
+    tmpl = """exec lxc-start --name %(container_name)s -f %(lxc_config_path)s -- su --preserve-environment --shell /bin/bash -c "cd /app;source /app/env.sh; exec %(cmd)s" %(user)s"""
+    return tmpl % vars()
+
+
+class LucidDeployer(ContainedDeployer):
+    mountpoints = ('/app',
+                   '/bin',
+                   '/dev',
+                   '/etc',
+                   '/lib',
+                   '/lib64',
+                   '/opt',
+                   '/usr',
+                   '/proc',
+                   '/sys',
+                   '/dev/pts',
+                   '/dev/shm')
+
+    lxc_config_template = 'lucid.lxc'
+
+
+class PreciseDeployer(ContainedDeployer):
+    mountpoints = ('/app',
+                   '/bin',
+                   '/dev',
+                   '/etc',
+                   '/lib',
+                   '/lib64',
+                   '/opt',
+                   '/usr',
+                   '/proc',
+                   '/run',
+                   '/sys',
+                   '/dev/pts',)
+                   #'/run/shm')
+
+    lxc_config_template = 'precise.lxc'
+
+
 @task
 def configure_proc(release_name, proc, port, user='nobody', use_syslog=False,
                    contain=False):
-
-    d = Deployment(release_name, proc, port, user, use_syslog, contain)
-    d.run()
+    if not contain:
+        d = Deployer
+    else:
+        issue = sudo('cat /etc/issue')
+        if '10.04' in issue:
+            d = LucidDeployer
+        elif '12.04' in issue:
+            d = PreciseDeployer
+        else:
+            raise ValueError("Could not determine deployer type for %s" %
+                             issue)
+    d(release_name, proc, port, user, use_syslog).run()
 
 
 @task
@@ -265,15 +370,14 @@ def build_contained_uptests_command(proc_path, proc, host, port, user):
     Build the command string for uptesting the given proc inside its lxc
     container.
     """
-    tmpl = """lxc-start --name %(procname)s-uptest -f %(lxc_config)s -- su --preserve-environment -c "cd /app;source /app/env.sh; exec /uptester %(folder)s %(host)s %(port)s" %(user)s"""
-    return tmpl % {
-        'procname': posixpath.basename(proc_path),
-        'lxc_config': posixpath.join(proc_path, 'proc.lxc'),
+    cmd = "/uptester %(folder)s %(host)s %(port)s" % {
         'folder': posixpath.join('/app/uptests', proc),
         'host': host,
         'port': port,
-        'user': user,
     }
+    container_name = posixpath.basename(proc_path) + '-uptest'
+    lxc_config_path = posixpath.join(proc_path, 'proc.lxc')
+    return build_container_cmd(cmd, user, container_name, lxc_config_path)
 
 
 def build_uncontained_uptests_command(release_path, proc_path, proc, host,
